
(define-module povray
  (use srfi-1)
  (use util.list)
  (export-all))
(select-module povray)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; utilities

(define (pov-escape-name str)
  (regexp-replace-all
   #/-/
   (regexp-replace-all
    #/[^-\w]/ str
    (lambda (m) (format "_%2X" (char->integer (string-ref (m 0) 0)))))
   "_"))

(define (keyword+object x)
  (list (class-keyword x) x))

(define (propq pl key)
  (let lp ((l pl))
    (if (null? l)
      #f
      (let ((rest (cdr l)))
        (if (eq? (car l) key)
          rest
          (lp (cdr rest)))))))

(define (get-key key ls . opt)
  (let loop ((ls ls))
    (if (pair? ls)
      (if (eq? key (car ls))
        (cadr ls)
        (loop (cadr ls)))
      (get-optional opt #f))))

(define (slot-name s) (if (pair? s) (car s) s))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; syntax

(define-syntax let-pov
  (syntax-rules ()
    ((_ ((name value) ...) . body)
     (let ((name (pov-name value 'name)) ...)
       . body))))

(define-syntax let-pov*
  (syntax-rules ()
    ((_ ((name value) ...) . body)
     (let* ((name (pov-name value 'name)) ...)
       . body))))

(define-class <pov-object> ()
  ((name :init-keyword :name)
   (transforms :init-keyword :transforms :init-value '())
   (attributes :init-keyword :attributes)))

(define-macro (define-pov-object constructor class-name slots . rest)
  (let ((ls (gensym)) (keys (gensym)))
    `(begin
       (define-class ,class-name ,slots ,@rest)
       (define (,(car constructor) . ,ls)
         (let-optionals* ,ls (,@(cdr constructor) . ,keys)
           (apply make ,class-name
                  ,@(append-map (lambda (x)
                                  (let ((s (slot-name x)))
                                    (list (get-key :init-keyword (assq-ref slots s '())
                                                   (make-keyword s))
                                          s)))
                                (cdr constructor))
                  (append-map keyword+object ,keys)))))))

(define-class <pov-attribute> (<pov-object>)
  ((value :init-keyword :value :init-value #f)))

(define-macro (define-pov-attribute constructor class-name)
  (let ((ls (gensym)) (keys (gensym)))
    `(begin
       (define-class ,class-name (<pov-attribute>) ())
       (define (,constructor . ,ls)
         (make ,class-name :value (get-optional ,ls #f))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; classes

(define-pov-attribute ambient <pov-ambient>)
(define-pov-attribute angle <pov-angle>)
(define-pov-attribute assumed-gamma <pov-assumed-gamma>)
(define-pov-attribute brilliance <pov-brilliance>)
(define-pov-attribute diffuse <pov-diffuse>)
(define-pov-attribute metallic <pov-metallic>)
(define-pov-attribute open <pov-open>)
(define-pov-attribute orthographic <pov-orthographic>)
(define-pov-attribute phong <pov-phong>)
(define-pov-attribute phong-size <pov-phong-size>)
(define-pov-attribute reflection <pov-reflection>)
(define-pov-attribute ripples <pov-ripples>)
(define-pov-attribute roughness <pov-roughness>)
(define-pov-attribute specular <pov-specular>)
(define-pov-attribute sturm <pov-sturm>)
(define-pov-attribute tolerance <pov-tolerance>)
(define-pov-attribute turbulence <pov-turbulence>)

(define-class <pov-csg> (<pov-object>) ())

(define-pov-object (union) <pov-union> (<pov-csg>) ())
(define-pov-object (difference) <pov-difference> (<pov-csg>) ())
(define-pov-object (intersection) <pov-intersection> (<pov-csg>) ())
(define-pov-object (merge) <pov-merge> (<pov-csg>) ())

(define-pov-object (camera) <pov-camera> (<pov-object>) ())

(define-pov-object (point x y z) <pov-point> (<pov-object>)
  ((x :init-keyword :x :init-value 0)
   (y :init-keyword :y :init-value 0)
   (z :init-keyword :z :init-value #f)))

(define-pov-object (%color r g b) <pov-color> (<pov-object>)
  ((r :init-keyword :r :init-value 0)
   (g :init-keyword :g :init-value 0)
   (b :init-keyword :b :init-value 0)))

(define-pov-object (background) <pov-background> (<pov-object>) ())
(define-pov-object (finish) <pov-finish> (<pov-object>) ())
(define-pov-object (normal) <pov-normal> (<pov-object>) ())
(define-pov-object (pigment) <pov-pigment> (<pov-object>) ())
(define-pov-object (text) <pov-text> (<pov-object>) ())
(define-pov-object (texture) <pov-texture> (<pov-object>) ())

(define-class <pov-valueless> (<pov-object>) ())
(define-class <pov-spline> (<pov-valueless>) ())
(define-class <pov-sweep> (<pov-valueless>) ())
(define-pov-object (b-spline) <pov-b-spline> (<pov-spline>) ())
(define-pov-object (bezier-spline) <pov-bezier-spline> (<pov-spline>) ())
(define-pov-object (cubic-spline) <pov-cubic-spline> (<pov-spline>) ())
(define-pov-object (linear-spline) <pov-linear-spline> (<pov-spline>) ())
(define-pov-object (conic-sweep) <pov-conic-sweep> (<pov-sweep>) ())
(define-pov-object (linear-sweep) <pov-linear-sweep> (<pov-sweep>) ())

(define-pov-object (checker color1 color2) <pov-checker> (<pov-object>)
  ((color1 :init-keyword :color1)
   (color2 :init-keyword :color2)))

(define-pov-object (lathe spline points) <pov-lathe> (<pov-object>)
  ((spline :init-keyword :spline)
   (points :init-keyword :points)))

(define-pov-object (sor points) <pov-sor> (<pov-object>)
  ((points :init-keyword :points)))

(define-pov-object (prism sweep spline height1 height2 points) <pov-prism> (<pov-object>)
  ((sweep :init-keyword :sweep)
   (spline :init-keyword :spline)
   (height1 :init-keyword :height1)
   (height2 :init-keyword :height2)
   (points :init-keyword :points)))

(define-pov-object (ttf font text thickness offset) <pov-ttf> (<pov-object>)
  ((font :init-keyword :font)
   (text :init-keyword :text)
   (thickness :init-keyword :thickness)
   (offset :init-keyword :offset)))

(define-pov-object (location) <pov-location> (<pov-object>)
  ((point :init-keyword :point)))
(define-pov-object (look-at) <pov-look-at> (<pov-object>)
  ((point :init-keyword :point)))

(define-pov-object (light-source location) <pov-light-source> (<pov-object>)
  ((location :init-keyword :location)))

(define-pov-object (box lower-left upper-right) <pov-box> (<pov-object>)
  ((lower-left :init-keyword :lower-left :init-value (point 0 0 0))
   (upper-right :init-keyword :upper-right :init-value (point 0 0 0))))

(define-pov-object (cone center1 radius1 center2 radius2) <pov-cone> (<pov-object>)
  ((center1 :init-keyword :center1 :init-value (point 0 0 0))
   (radius1 :init-keyword :radius1 :init-value 1)
   (center2 :init-keyword :center2 :init-value (point 0 0 0))
   (radius2 :init-keyword :radius2 :init-value 1)))

(define-pov-object (cylinder center1 center2 radius) <pov-cylinder> (<pov-object>)
  ((center1 :init-keyword :center1 :init-value (point 0 0 0))
   (center2 :init-keyword :center2 :init-value (point 0 0 0))
   (radius :init-keyword :radius :init-value 1)))

(define-pov-object (plane normal displacement) <pov-plane> (<pov-object>)
  ((normal :init-keyword :normal :init-value (point 0 0 0))
   (displacement :init-keyword :displacement :init-value 0)))

(define-pov-object (torus radius1 radius2) <pov-torus> (<pov-object>)
  ((radius1 :init-keyword :radius1 :init-value 1)
   (radius2 :init-keyword :radius2 :init-value 0)))

(define-pov-object (sphere location radius) <pov-sphere> (<pov-object>)
  ((location :init-keyword :location :init-value (point 0 0 0))
   (radius :init-keyword :radius :init-value 1)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; colors

;; can be an internal modifier or a functional updater
(define (color . ls)
  (define (return val tail)
    (if (and (pair? tail) (is-a? (car tail) <pov-object>))
      (let* ((res (copy (car tail)))
             (attrs (slot-ref res 'attributes))
             (tx (propq attrs :texture)))
        (if tx
          (set-car! tx (texture (pigment val)))
          (set-cdr! (last-pair attrs) (list (texture (pigment val)))))
        res)
      val))
  (cond 
    ((is-a? (car ls) <pov-color>)
     (return (car ls) (cdr ls)))
    (else
     (return (apply %color ls) (drop ls 3)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; constants

(define x (point 1 0 0))
(define y (point 0 1 0))
(define z (point 0 0 1))
(define origin (point 0 0 0))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; arithmetic

(define-method object-+ ((a <pov-point>) (b <pov-point>))
  (point (+ (slot-ref a 'x) (slot-ref b 'x))
         (+ (slot-ref a 'y) (slot-ref b 'y))
         (+ (slot-ref a 'z) (slot-ref b 'z))))

(define-method object-+ ((a <pov-point>) (b <number>))
  (point (+ (slot-ref a 'x) b) (+ (slot-ref a 'y) b) (+ (slot-ref a 'z) b)))

(define-method object-+ ((a <number>) (b <pov-point>))
  (object-+ b a))

(define-method object-- ((a <pov-point>) (b <pov-point>))
  (point (- (slot-ref a 'x) (slot-ref b 'x))
         (- (slot-ref a 'y) (slot-ref b 'y))
         (- (slot-ref a 'z) (slot-ref b 'z))))

(define-method object-- ((a <pov-point>) (b <number>))
  (point (- (slot-ref a 'x) b) (- (slot-ref a 'y) b) (- (slot-ref a 'z) b)))

(define-method object-- ((a <number>) (b <pov-point>))
  (point (- a (slot-ref b 'x)) (- a (slot-ref b 'y)) (- a (slot-ref b 'z))))

(define-method object-* ((a <pov-point>) (b <pov-point>))
  (point (* (slot-ref a 'x) (slot-ref b 'x))
         (* (slot-ref a 'y) (slot-ref b 'y))
         (* (slot-ref a 'z) (slot-ref b 'z))))

(define-method object-* ((a <pov-point>) (b <number>))
  (point (* (slot-ref a 'x) b) (* (slot-ref a 'y) b) (* (slot-ref a 'z) b)))

(define-method object-* ((a <number>) (b <pov-point>))
  (object-* b a))

(define-method object-* ((a <pov-color>) (b <number>))
  (color (* (slot-ref a 'r) b) (* (slot-ref a 'g) b) (* (slot-ref a 'b) b)))

(define-method object-* ((a <number>) (b <pov-color>))
  (object-* b a))

(define-method object-/ ((a <pov-point>) (b <number>))
  (point (/ (slot-ref a 'x) b) (/ (slot-ref a 'y) b) (/ (slot-ref a 'z) b)))

(define-method object-/ ((a <number>) (b <pov-point>))
  (point (/ a (slot-ref b 'x)) (/ a (slot-ref b 'y)) (/ a (slot-ref b 'z))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; methods

(define-method copy (x) x)

(define-method copy ((x <pair>)) (cons (copy (car x)) (copy (cdr x))))

(define-method copy ((x <vector>))
  (let ((res (make-vector (vector-length x))))
    (do ((i (- (vector-length x) 1) (- i 1)))
        ((negative? i) res)
      (vector-set! res i (copy (vector-ref x i))))))

(define-method copy ((x <pov-object>))
  (let ((res (make (class-of x))))
    (for-each
     (lambda (s) (if (slot-bound? x s) (slot-set! res s (copy (slot-ref x s)))))
     (map slot-name (class-slots (class-of x))))
    res))

(define-method pov-name (x name)
  x)

(define-method pov-name ((x <pov-object>) name)
  (slot-set! x 'name name)
  x)

(define-method initialize ((x <pov-object>) initargs)
  (next-method)
  (slot-set! x 'attributes initargs))

(define-method base-attributes ((x <pov-object>))
  (let ((bad-slots (map slot-name (class-slots <pov-object>))))
    (let loop ((ls (map slot-name (remove (lambda (x)
                                            (and (pair? x)
                                                 (get-key :positional (cdr x))))
                                          (class-slots (class-of x)))))
               (res '()))
      (cond ((null? ls) (reverse res))
            ((memq (car ls) bad-slots) (loop (cdr ls) res))
            ((slot-bound? x (car ls))
             (loop (cdr ls) (cons (slot-ref x (car ls)) res)))
            (else (loop (cdr ls) res))))))

(define-method extra-attributes ((x <pov-object>))
  (let loop ((ls (slot-ref x 'attributes)) (res '()))
    (cond
      ((null? ls) (reverse res))
      ((keyword? (car ls))
       (if (slot-exists? x (string->symbol (keyword->string (car ls))))
         (loop (cddr ls) res)
         (loop (cddr ls) (cons (cadr ls) (cons (car ls) res)))))
      ((slot-exists? x (string->symbol (keyword->string (class-keyword (car ls)))))
       (loop (cdr ls) res))
      (else (loop (cdr ls) (cons (car ls) (cons (class-keyword (car ls)) res)))))))

(define-method pov->string-chain-slot (x) #f)
(define-method pov->string-chain-slot ((x <pov-location>)) 'point)
(define-method pov->string-chain-slot ((x <pov-look-at>)) 'point)

(define-method class-keyword ((x <pov-object>))
  (make-keyword
   (regexp-replace
    #/-/
    (regexp-replace #/^<pov-([^<>]*)>$/
                    (x->string (slot-ref (class-of x) 'name))
                    "\\1")
    "_")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; transformations

(define-method scale (size x) x)

(define-method scale (size (x <pov-object>))
  (let ((res (copy x)))
    (slot-set! res 'transforms (cons `(scale ,size) (slot-ref res 'transforms)))
    res))

(define-method scale ((size <number>))
  (lambda (x) (scale size x)))

(define-method translate (pt x) x)

(define-method translate (pt (x <pov-object>))
  (let ((res (copy x)))
    (slot-set! res 'transforms (cons `(translate ,pt) (slot-ref res 'transforms)))
    res))

(define-method translate ((pt <pov-point>))
  (lambda (x) (translate pt x)))

(define-method rotate (ang x) x)

(define-method rotate (ang (x <pov-object>))
  (let ((res (copy x)))
    (slot-set! res 'transforms (cons `(rotate ,ang) (slot-ref res 'transforms)))
    res))

(define-method rotate ((ang <pov-point>))
  (lambda (x) (rotate ang x)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; printing

(define (pov->string x env)
  (if (and (slot-exists? x 'name) (slot-bound? x 'name) (slot-ref x 'name))
    (let ((name (slot-ref x 'name)))
      (hash-table-put! env name x)
      (pov-escape-name (x->string name)))
    (%pov->string x env)))

(define-method %pov->string (x (env <hash-table>))
  (write-to-string x))

(define-method %pov->string ((x <keyword>) (env <hash-table>))
  (x->string x))

(define-method %pov->string ((x <null>) (env <hash-table>))
  "")

(define-method %pov->string ((x <list>) (env <hash-table>))
  (string-join (cons (number->string (length x))
                     (map (cute pov->string <> env) x))
               ", "))

(define-method %pov->string ((x <vector>) (env <hash-table>))
  (string-append "<" (string-join (map (cute pov->string <> env)
                                       (vector->list x))
                                  ",")
                 ">"))

(define-method %pov->string ((x <pov-object>) (env <hash-table>))
  (let ((s (pov->string-chain-slot x))
        (sep (if (is-a? x <pov-csg>) "\n" " ")))
    (if s
      (%pov->string (slot-ref x s) env)
      (let ((base (base-attributes x))
            (extra (extra-attributes x))
            (transforms (slot-ref x 'transforms)))
        (string-append
         "{" (string-join (map (cut pov->string <> env)
                               base) sep)
         (if (and (pair? base) (pair? extra)) sep "")
         (string-join (map (cut pov->string <> env)
                           extra) sep)
         (if (pair? transforms)
           (string-join (map (lambda (x) (format " ~A ~A" (car x)
                                            (pov->string (cadr x) env)))
                             (reverse transforms)) "")
           "")
         "}")))))

(define-method %pov->string ((x <pov-attribute>) (env <hash-table>))
  (let ((val (slot-ref x 'value)))
    (if val
      (%pov->string val env)
      "")))

(define-method %pov->string ((x <pov-point>) (env <hash-table>))
  (let ((z (slot-ref x 'z)))
    (if z
      (format "<~A,~A,~A>" (slot-ref x 'x) (slot-ref x 'y) (slot-ref x 'z))
      (format "<~A,~A>" (slot-ref x 'x) (slot-ref x 'y)))))

(define-method %pov->string ((x <pov-color>) (env <hash-table>))
  (format "rgb<~A,~A,~A>" (slot-ref x 'r) (slot-ref x 'g) (slot-ref x 'b)))

(define-method %pov->string ((x <pov-ttf>) (env <hash-table>))
  (format "~S ~S ~S, ~A" (slot-ref x 'font) (slot-ref x 'text)
          (slot-ref x 'thickness) (pov->string (slot-ref x 'offset) env)))

(define-method %pov->string ((x <pov-valueless>) (env <hash-table>))
  (keyword->string (class-keyword x)))

(define-method %pov->string ((x <pov-checker>) (env <hash-table>))
  (format "~A, ~A" (pov->string (slot-ref x 'color1) env)
          (pov->string (slot-ref x 'color2) env)))

(define (scene . ls)
  (let* ((env (make-hash-table))
         (dummy (make-hash-table))
         (res (string-join (map (lambda (x)
                                  (format "~A ~A"
                                          (keyword->string (class-keyword x))
                                          (pov->string x env)))
                                ls) "\n")))
    (string-append
     (string-join (map (lambda (name)
                         (format "#define ~A ~A"
                                 (pov-escape-name (x->string name))
                                 (%pov->string (hash-table-get env name) dummy)))
                       (hash-table-keys env))
                  "\n")
     (if (pair? (hash-table-keys env)) "\n\n" "")
     res
     "\n")))

(provide "povray")
